#include "common.h"
#include "MainWindow.h"
#include "CommanderWeb.h"
#include "XLFileUtil.h"
#include "XLStringUtil.h"
#include "XLSoundPlay.h"
#include "XLSocket.h"

CommanderWeb::CommanderWeb()
{
}
CommanderWeb::~CommanderWeb()
{
}
void CommanderWeb::Create()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	//secretがない場合は新規に作る
	const std::string secret = MainWindow::m()->Config.Get("local_secret","");
	if (secret.empty() )
	{
		MainWindow::m()->Config.Set("local_secret",XLStringUtil::uuid());
		MainWindow::m()->Config.overrideSave();
	}
}


bool CommanderWeb::checkRemoteAccess(const XLHttpHeader&httpHeaders,unsigned long ip,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	const auto headers = httpHeaders.getHeaderPointer();
	const auto key = mapfind(headers,"x-fhcsecret","");
	const std::string localsecret = MainWindow::m()->Config.Get("local_secret","") ;
	const std::string localsecret_sha1 = XLStringUtil::sha1( localsecret );
	if ( localsecret_sha1 != key)
	{//アクセスキーが一致しません
		*result = WEBSERVER_RESULT_TYPE_FORBIDDEN;
		*responsString = "NG\r\nsecret no match!!";
		return false;
	}

	//LAN IP からのアクセスのみ許可します。
	static unsigned long local10Start = inet_addr("10.0.0.0");
	static unsigned long local10End = inet_addr("10.255.255.255");
	static unsigned long local172Start = inet_addr("172.16.0.0");
	static unsigned long local172End = inet_addr("172.31.255.255");
	static unsigned long local192Start = inet_addr("192.168.0.0");
	static unsigned long local192End = inet_addr("192.168.255.255");
	static unsigned long local127 = inet_addr("127.0.0.1");
	if (ip > local10Start && ip < local10End )
	{
		return true;
	}
	if (ip > local172Start && ip < local172End )
	{
		return true;
	}
	if (ip > local192Start && ip < local192End )
	{
		return true;
	}
	if (ip == local127)
	{
		return true;
	}

	*result = WEBSERVER_RESULT_TYPE_FORBIDDEN;
	*responsString = "NG\r\nLAN ONLY!! you need access 10.0.0.0/8 or 172.16.0.0/16 or 192.168.0.0/16 or 127.0.0.1";
	return false;
}

bool CommanderWeb::checkLocalAccess(const XLHttpHeader&httpHeaders,unsigned long ip,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	//LCOAL IP からのアクセスのみ許可します。
	static unsigned long local127 = inet_addr("127.0.0.1");
	if (ip == local127)
	{
		return true;
	}

	*result = WEBSERVER_RESULT_TYPE_FORBIDDEN;
	*responsString = "NG\r\nLOCAL ONLY!! you need access 127.0.0.1";
	return false;
}


//リモート実行
std::string CommanderWeb::ProcExecute(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	const auto post = httpHeaders.getPostPointer();
	const auto command = mapfind(post,"command","");
	const auto args = mapfind(post,"args","");

#if _MSC_VER
	ShellExecute( NULL , NULL , command.c_str() , args.c_str() , NULL , SW_SHOW);
#else
	
#endif
	return "OK";
}

//リモートスピーカー
std::string CommanderWeb::ProcSoundPlay(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	const auto get = httpHeaders.getGetPointer();
	const auto files = httpHeaders.getFilesPointer();
	if ( files->begin() == files->end() )
	{
		*result = WEBSERVER_RESULT_TYPE_FORBIDDEN;
		return "BAD\r\nno file";
	}
	const std::string tempdir = XLFileUtil::getTempDirectory("fhc_commander");
	std::string tempfilename = XLStringUtil::pathseparator(tempdir + "/" + num2str(time(NULL)) + "_" + num2str(rand()));

	const auto type = mapfind(get,"type","");
	auto firstfileIT = files->begin();
	if ( type == "mp3" ) 
	{//mp3
		tempfilename = tempfilename + ".mp3";
	}
	else if ( type == "wav" ) 
	{//wav
		tempfilename = tempfilename + ".wav";
	}
	else 
	{
		*result = WEBSERVER_RESULT_TYPE_FORBIDDEN;
		return "BAD\r\ncan not support type";
	}

	XLFileUtil::write(tempfilename , firstfileIT->second->data);
	XLSoundPlay play;
	play.play(tempfilename);

	XLFileUtil::del(tempfilename);
	return "OK";
}

//webrootからのパスを得る static
std::string GetWebRootPath(const std::string& path) 
{
	return MainWindow::m()->GetConfigBasePath("webroot/" + path);
}
//tpl読み込み static
std::string OpenTPL(const XLHttpHeader& httpHeaders,const std::string& tplname)  
{
	const std::string filename = GetWebRootPath(tplname);

	const auto header = httpHeaders.getHeaderPointer();
	const auto useragent = mapfind(header,"user-agent");

	if ( strstr(useragent.c_str(),"WebKit" ))
	{
		return XLStringUtil::replace_low( XLFileUtil::cat(filename) ,"%BCSS%" , "-webkit-");
	}

	if ( strstr(useragent.c_str(),"Firefox"  ))
	{
		return XLStringUtil::replace_low( XLFileUtil::cat(filename) ,"%BCSS%" , "-moz-");
	}

	if ( strstr(useragent.c_str(),"Opera"  ))
	{
		return XLStringUtil::replace_low( XLFileUtil::cat(filename) ,"%BCSS%" , "-o-");
	}

	if ( strstr(useragent.c_str(),"MSIE" ))
	{
		return XLStringUtil::replace_low( XLFileUtil::cat(filename) ,"%BCSS%" , "-ms-");
	}

	return XLStringUtil::replace_low( XLFileUtil::cat(filename) ,"%BCSS%" , "");
}

//自分に外から接続するIPアドレスを見つけます
std::string GetMyHost()
{
	const int port = MainWindow::m()->Config.GetInt("httpd__port", 15551 );
	const auto interfaceMap = XLSocket::getIPAddressMap();
	auto it = interfaceMap.end();
#if _MSC_VER
	if ( interfaceMap.size() >= 1 )
	{//とりあえず最初に見つかった稼働しているものを使う.
		it = interfaceMap.begin();
	}
#else
	if ( interfaceMap.size() >= 1 )
	{
		it = interfaceMap.find("wlan0");
		if (it != interfaceMap.end() && !it->second.empty() )
		{//無線があれば無線を優先的に使う
		}
		else
		{
			it = interfaceMap.find("eth0");
			if (it != interfaceMap.end())
			{//有線ケーブル
			}
			else
			{//わかんないので最初に見つかった奴
				it = interfaceMap.begin();
			}
		}
	}
#endif
	if (it == interfaceMap.end())
	{//何もネットデバイスがない
		NOTIFYLOG("ネットワークデバイスがありませんでした。 127.0.0.1 にします.");
		return "127.0.0.1:" + num2str(port);
	}
	else
	{//見つかったものにする
		NOTIFYLOG("IPアドレス:" + it->second + "(" + it->first + ")を利用します.");
		return it->second + ":" + num2str(port) ;
	}
}

//about
std::string CommanderWeb::About_Index(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	return OpenTPL(httpHeaders,"about.tpl");
}

//情報閲覧画面トップ
std::string CommanderWeb::Local_Index(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	const std::string html = OpenTPL(httpHeaders,"local.tpl");

	
	std::string jsonstring;
	jsonstring += std::string(",\"") + XLStringUtil::jsonescape("local_host") + "\": " 
		+ "\"" + XLStringUtil::jsonescape(GetMyHost()) + "\"";
	jsonstring += std::string(",\"") + XLStringUtil::jsonescape("local_secret") + "\": " 
		+ "\"" + XLStringUtil::jsonescape(MainWindow::m()->Config.Get("local_secret","")) + "\"";

	jsonstring = std::string("{ \"status\": \"ok\" ") + jsonstring + "}";
	return XLStringUtil::replace_low(html,"%LOCAL_VALUE%", jsonstring );
}

//新規にsecretを割り当てる
std::string CommanderWeb::local_secret_update(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	_USE_WINDOWS_ENCODING;

	MainWindow::m()->Config.Set("local_secret",XLStringUtil::uuid());
	MainWindow::m()->Config.overrideSave();

	std::string jsonstring;
	jsonstring += std::string(",\"") + XLStringUtil::jsonescape("local_secret") + "\": " 
		+ "\"" + XLStringUtil::jsonescape(MainWindow::m()->Config.Get("local_secret","")) + "\"";

	jsonstring = std::string("{ \"status\": \"ok\" ") + jsonstring + "}";
	return _A2U(jsonstring);
}

std::string CommanderWeb::local_fhc_search(const XLHttpHeader&httpHeaders,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	_USE_WINDOWS_ENCODING;

	std::string jsonstring;
	int foundcount = 1;

	//UPNP通信して聞きます。
	XLSocket socket;
	socket.CreateUDP(5); //serverに応答を返さないといけないので、 5秒だけ聞きます
	socket.SetSockopt(SOL_SOCKET,SO_BROADCAST, 1);

	struct sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_port = htons(1900);
	addr.sin_addr.s_addr = inet_addr("239.255.255.250");

	const std::string msearch = 
			"M-SEARCH * HTTP/1.1\r\n" 
			"HOST: 239.255.255.250:1900\r\n" 
			"MAN: \"ssdp:discover\"\r\n" 
			"ST: upnp:rootdevice:rti-giken:naichichi2\r\n"
			"\r\n"
			;
	socket.SendTo(msearch.c_str() , msearch.size() , 0 , &addr , sizeof(addr) );
	while(1)
	{
		struct sockaddr_in senderinfo;
		char buffer[8192];
		int fromlen = sizeof(senderinfo);
		int size = socket.RecvFrom(buffer, sizeof(buffer)-1,0,&senderinfo,&fromlen);
		if (size < 0)
		{
			break;
		}
		buffer[size] = 0;
		DEBUGLOG("UPNP recvfrom: from IP " << inet_ntoa(senderinfo.sin_addr) << " ("<< buffer<<")\n" );

		XLHttpHeader header;
		if ( ! header.Parse( buffer ,size ) )
		{
			DEBUGLOG("HTTPDヘッダーとしてパース出来ません");
			continue;
		}
		//server ヘッダに naichichi2 と入っていたら俺です。
		if ( header.getAt("server").find("naichichi2") == std::string::npos )
		{
			DEBUGLOG("SERVERアトリビュートが naichichi2 ではありません 値:" << header.getAt("server"));
			continue;
		}

		//本当は locationヘッダの先にあるxmlに書いてある presentationURL をとってくるのが正攻法なんですが、
		//めんどいので、locationヘッダーからURLを作ります。
		const std::string location = header.getAt("location");
		const std::vector<std::string> vec = XLStringUtil::split_vector("/",location);
		if ( vec.size() <= 3 )
		{
			DEBUGLOG("locationアトリビュートが不明な値です 値:" << header.getAt("location"));
			continue;
		}

		jsonstring += std::string(",\"") + XLStringUtil::jsonescape("found_"+num2str(foundcount) + "_host") + "\": " 
			+ "\"" + XLStringUtil::jsonescape(vec[2]) + "\"";
		foundcount++;
	}

	jsonstring = std::string("{ \"status\": \"ok\" ") + jsonstring + "}";
	return _A2U(jsonstring);
}

//webからアクセスがあったときに呼ばれます。
bool CommanderWeb::WebAccess(const std::string& path,const XLHttpHeader& httpHeaders,unsigned long ip,WEBSERVER_RESULT_TYPE* result,std::string* responsString)
{
	if (path == "/local/")
	{
		if (!checkLocalAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_HTML;
		*responsString = Local_Index(httpHeaders,result,responsString);
		return true;
	}
	if (path == "/about/")
	{
		if (!checkLocalAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_HTML;
		*responsString = About_Index(httpHeaders,result,responsString);
		return true;
	}
	else if (path == "/local/local_secret_update")
	{
		if (!checkLocalAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_JSON;
		*responsString = local_secret_update(httpHeaders,result,responsString);
		return true;
	}
	else if (path == "/local/local_fhc_search")
	{
		if (!checkLocalAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_JSON;
		*responsString = local_fhc_search(httpHeaders,result,responsString);
		return true;
	}
	else if (path == "/api/execute")
	{
		if (!checkRemoteAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_HTML;
		*responsString = ProcExecute(httpHeaders,result,responsString);
		return true;
	}
	else if (path == "/api/play")
	{
		if (!checkRemoteAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_HTML;
		*responsString = ProcSoundPlay(httpHeaders,result,responsString);
		return true;
	}
	else if (path == "/api/remote_auth")
	{
		if (!checkRemoteAccess(httpHeaders,ip,result,responsString)) return true;

		*result = WEBSERVER_RESULT_TYPE_OK_HTML;
		*responsString = "OK";
		return true;
	}
	return false;
}

