﻿//GPLでライセンスされています。(他のライセンス形態を望むならばお問い合わせください)
#include "common.h"
#include "Config.h"
#include "MainWindow.h"
#include "XLStringUtil.h"
#include "XLFileUtil.h"

Config::Config()
{
}
Config::~Config()
{
}

void Config::Create(const std::string& config)
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	this->loadConfig(config);
}

bool Config::overrideSave()
{
	if ( ! this->ConfigFullpath.empty() )
	{
		this->saveConfig(this->ConfigFullpath);
	}
	return true;
}



bool Config::loadConfig(const std::string & configFilename)
{
	boost::unique_lock<boost::mutex> al(this->lock);
	_USE_WINDOWS_ENCODING;
	this->ConfigData.clear();

	FILE * fp = fopen( XLStringUtil::pathseparator(configFilename).c_str() ,"rb");
	if (fp == NULL)
	{
		throw XLException("config " + configFilename + " を読み込めませんでした。");
	}
	char readbuffer[MAX_PATH];

	while( !feof(fp) )
	{
		readbuffer[0] = 0;
		if ( fgets(readbuffer , MAX_PATH-1 , fp) == NULL ) break;
		readbuffer[MAX_PATH-1] = 0;
		if (readbuffer[0] == 0) continue;
		const std::string str = _U2A(readbuffer);
		const char* buf = str.c_str();

		if ( this->IsComment(buf) ) continue;

		const char * eq = strchr( buf , '=');
		if (eq == NULL) continue;

		std::string key = XLStringUtil::chop(  std::string(buf , 0 , eq - buf) ) ;
		std::string value = XLStringUtil::chop( std::string(eq + 1) ) ;

		this->ConfigData[key] = value;
	}
	fclose(fp);

	this->ConfigFullpath = configFilename;
	return true;
}

bool Config::saveConfig(const std::string & configFilename)
{
	boost::unique_lock<boost::mutex> al(this->lock);
	_USE_WINDOWS_ENCODING;

	std::map<std::string,std::string>	localSetting = this->ConfigData;

	std::string tmpfilename = configFilename + ".tmp";
	FILE * wfp = fopen( XLStringUtil::pathseparator(tmpfilename).c_str() ,"wb");
	if (wfp == NULL)
	{
		throw XLException("設定保存用の一時ファイル" + tmpfilename + "を作成できませんでした。");
	}
	FILE * fp = fopen( XLStringUtil::pathseparator(configFilename).c_str() ,"rb");
	if (fp != NULL)
	{
			char readbuffer[MAX_PATH];
			int kakaline = 0;

			while( !feof(fp) )
			{
				readbuffer[0] = 0;
				if ( fgets(readbuffer , MAX_PATH-1 , fp) == NULL ) break;
				readbuffer[MAX_PATH-1] = 0;
				if (readbuffer[0] == 0) continue;

				const std::string str = _U2A(readbuffer);
				const char* buf = str.c_str();
				if ( (buf[0] == '\n' && buf[1] == '\0') || (buf[0] == '\r' && buf[1] == '\n' && buf[2] == '\0'))
				{
					kakaline++;
					if(kakaline < 5)
					{//複数行連続する改行は消す
						fprintf(wfp , "\r\n" );
					}
					continue;
				}
				kakaline=0;

				if ( this->IsComment(buf) )
				{
					fprintf(wfp , "%s" , _A2U(buf));
					continue;
				}

				const char * eq = strchr( buf , '=');
				if (eq == NULL)
				{//ゴミデータなので無視する
//					fprintf(wfp , "%s" , _A2U(buf));
					continue;
				}

				const std::string key = XLStringUtil::chop( std::string(buf , 0 , eq - buf) ) ;
				const std::string value = XLStringUtil::chop( std::string(eq + 1) ) ;

				//このデータは読み込んでいますか?
				const auto i = localSetting.find(key);
				if (i == localSetting.end() )
				{
					//ない場合は削除
					continue;
				}
				if (value == (*i).second)
				{
					fprintf(wfp , "%s" , _A2U(buf));
					localSetting.erase(i);
					continue;
				}

				//書き換え.
				const std::string writekey = _A2U((*i).first);
				const std::string writevalue = _A2U((*i).second);
				if ( writekey.empty() )
				{
					continue;
				}
				fprintf(wfp , "%s=%s\r\n" , writekey.c_str()  , writevalue.c_str() );
				localSetting.erase(i);
			}
	}
	//まだ書き込んで値は追加.
	{
		for(std::map<std::string,std::string>::iterator i = localSetting.begin() ; i != localSetting.end() ; ++i)
		{
			const std::string writekey = _A2U((*i).first);
			const std::string writevalue = _A2U((*i).second);
			fprintf(wfp , "%s=%s\r\n" , writekey.c_str()  , writevalue.c_str() );
		}
	}
	if (fp) fclose(fp);
	if (wfp) fclose(wfp);

	XLFileUtil::copy(tmpfilename,configFilename);
	XLFileUtil::del( tmpfilename );

	return true;
}




void Config::Set(const std::string& key,const std::string& value ) 
{
	boost::unique_lock<boost::mutex> al(this->lock);
	this->ConfigData[key] = value;
}

void Config::Remove(const std::string& key ) 
{
	boost::unique_lock<boost::mutex> al(this->lock);

	auto it = this->ConfigData.find(key);

	if (it != this->ConfigData.end())
	{
		this->ConfigData.erase(it);
	}
}

std::string Config::Get(const std::string & key , const std::string & defaultValue) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	auto it = this->ConfigData.find(key);
	if (it == this->ConfigData.end())
	{
		return defaultValue;
	}
	return it->second;
}

double Config::GetDouble(const std::string & key , double defaultValue) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	auto it = this->ConfigData.find(key);
	if (it == this->ConfigData.end())
	{
		return defaultValue;
	}
	return atof(it->second.c_str());
}

int Config::GetInt(const std::string & key , int defaultValue) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	auto it = this->ConfigData.find(key);
	if (it == this->ConfigData.end())
	{
		return defaultValue;
	}
	return atoi(it->second.c_str());
}

bool Config::GetBool(const std::string & key , bool defaultValue) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	auto it = this->ConfigData.find(key);
	if (it == this->ConfigData.end())
	{
		return defaultValue;
	}
	return atoi(it->second.c_str()) != 0;
}

std::list<std::string> Config::FindGets(const std::string & prefix) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	std::list<std::string> r;
	for(auto it = this->ConfigData.begin() ; it != this->ConfigData.end() ; ++it )
	{
		if ( it->first.find(prefix) == 0 )
		{
			r.push_back(it->second);
		}
	}
	return r;
}

const std::map<std::string,std::string>Config::FindGetsToMap(const std::string & prefix,bool prefixtrim) const
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	std::map<std::string,std::string> r;
	for(auto it = this->ConfigData.begin() ; it != this->ConfigData.end() ; ++it )
	{
		if ( it->first.find(prefix) == 0 )
		{
			if (prefixtrim)
			{
				const std::string key = it->first.substr( prefix.size() );
				r.insert( std::pair<std::string,std::string>(key, it->second) );
			}
			else
			{
				r.insert( *it );
			}
		}
	}
	return r;
}

bool Config::ReplaceMap(const std::string& prefix,const std::map<std::string,std::string>& postData)
{
	boost::unique_lock<boost::mutex> al(this->lock);
	
	//まず該当するデータを消す
	for(auto it = this->ConfigData.begin()  , end = this->ConfigData.end();	it != end ; )
	{
		if ( it->first.find(prefix) == 0)
		{
			this->ConfigData.erase(it++);
		}
		else
		{
			++it;
		}
	}

	//もう一度データを突っ込む
	this->ConfigData.insert(postData.begin() , postData.end() );

	return true;
}



