﻿// HttpServer.cpp: HttpServer クラスのインプリメンテーション
//
//////////////////////////////////////////////////////////////////////

#include "common.h"
#include "HttpServer.h"
#include "MainWindow.h"
#include "XLHttpHeader.h"
#include "XLStringUtil.h"
#include "XLFileUtil.h"
#include "XLSoundPlay.h"




HttpWorker::HttpWorker(boost::asio::ip::tcp::socket* socket) : ConnectSocket(socket)
{
}
HttpWorker::~HttpWorker()
{
}


void HttpWorker::HTTP500()
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 500 Internal Server Error\r\n"
						"Server:naichichi2\r\n"
						"Connection: close\r\n"
						"\r\n"
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}
void HttpWorker::HTTP404()
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 404 Not Found\r\n"
						"Server:naichichi2\r\n"
						"Connection: close\r\n"
						"\r\n"
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}

void HttpWorker::HTTP403()
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 403 Forbidden\r\n"
						"Server:naichichi2\r\n"
						"Connection: close\r\n"
						"\r\n"
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}

void HttpWorker::HTTP403(const std::string& message)
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 403 Forbidden\r\n"
						"Server:naichichi2\r\n"
						"Connection: close\r\n"
						"Content-Length: " << message.size() << "\r\n"
						"\r\n"
						<< message
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}

void HttpWorker::HTTP302(const std::string& url)
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 302 Found\r\n"
						"Server:naichichi2\r\n"
						"Connection: close\r\n"
						"Location: " << XLStringUtil::clearcontrollcode(url) << "\r\n"
						"\r\n"
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}

void HttpWorker::HTTP200(const std::string& contents,const std::string& appendheaders,const std::map<std::string,std::string>& header,bool isKeepAlive)
{
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	respons_stream <<	"HTTP/1.0 200 OK\r\n"
						"Connection: close\r\n"
						"Server:naichichi2\r\n"
						"X-Frame-Options: SAMEORIGIN\r\n"
						;

	if (!appendheaders.empty())
	{
		respons_stream << appendheaders ;
	}

	respons_stream		<< "Content-Length: " << contents.size() << "\r\n"
						<< "\r\n"
						<< contents 
						<< std::flush;

	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
}

	
void HttpWorker::HTTP200SendFileContent(const std::string& realpath,const std::string& mime,const std::map<std::string,std::string>& header,bool isKeepAlive)
{
	//アクセスが許可されている拡張子か？
	ASSERT(!mime.empty());
	ASSERT(XLFileUtil::Exist(realpath));

	//更新されているか？
	const time_t unixtime = XLFileUtil::getfiletime(realpath);
	const auto ifmodifiedSinceIT = header.find("if-modified-since");
	if (ifmodifiedSinceIT != header.end())
	{
		time_t ifmodifiedSinceTime = XLStringUtil::HttpTimeToUnixTime( ifmodifiedSinceIT->second );
		if (ifmodifiedSinceTime >= unixtime)
		{
			boost::asio::streambuf respons;
			std::ostream respons_stream(&respons);
			if (isKeepAlive)
			{//keepaliveする
				respons_stream <<	"HTTP/1.0 304 Not Modified\r\n"
									"Connection: keep-alive\r\n"
									"Keep-Alive: timeout=5, max=100\r\n"
									;
			}
			else
			{//keepaliveしない
				respons_stream <<	"HTTP/1.0 304 Not Modified\r\n"
									"Connection: close\r\n"
									;
			}
			respons_stream <<	"Server:naichichi2\r\n"
								"X-Frame-Options: SAMEORIGIN\r\n"
								"Content-Length: 0\r\n"
								"\r\n";
			boost::system::error_code ec;
			boost::asio::write(*this->ConnectSocket,respons ,ec);
			return ;
;
		}
	}
	//ファイルを読み込む
	std::vector<char> filebinary = XLFileUtil::cat_b(realpath);

	//とりあえずさきっちゃだけ
    boost::asio::streambuf respons;
    std::ostream respons_stream(&respons);
	if (isKeepAlive)
	{//keepaliveする
		respons_stream <<	"HTTP/1.0 200 OK\r\n"
							"Connection: keep-alive\r\n"
							"Keep-Alive: timeout=5, max=100\r\n"
							;
	}
	else
	{//keepaliveしない
		respons_stream <<	"HTTP/1.0 200 OK\r\n"
							"Connection: close\r\n"
							;
	}
	respons_stream <<	"Server:naichichi2\r\n"
						"Last-Modified: " << XLStringUtil::UnixTimeToHttpTime( unixtime ) << "\r\n" <<
						"Content-Type: " << mime << "\r\n"
						<< std::flush;

	respons_stream <<	"Content-Length: " << filebinary.size() << "\r\n"
						"\r\n"
						<< std::flush;
	boost::system::error_code ec;
	boost::asio::write(*this->ConnectSocket,respons ,ec);
	boost::asio::write(*this->ConnectSocket,boost::asio::buffer(&filebinary[0],filebinary.size() ) ,ec);
	return ;
}

void HttpWorker::operator()()
{
#if _MSC_VER
    int32_t timeout = 5000;
    setsockopt(this->ConnectSocket->native(), SOL_SOCKET, SO_RCVTIMEO, (const char*)&timeout, sizeof(timeout));
    setsockopt(this->ConnectSocket->native(), SOL_SOCKET, SO_SNDTIMEO, (const char*)&timeout, sizeof(timeout));
#else
    struct timeval tv;
    tv.tv_sec  = 5; 
    tv.tv_usec = 0;         
    setsockopt(this->ConnectSocket->native(), SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
    setsockopt(this->ConnectSocket->native(), SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv));
#endif

	while(1)
	{
		try
		{
			if ( ! HttpTrans() )
			{
				break;
			}
		}
		catch(XLException& e)
		{
			ERRORLOG(e.getFullErrorMessage());
			break;
		}

		//パイプライン転送を行います
	}

	//切断
	this->ConnectSocket->close();
}

unsigned long HttpWorker::getRemoteIP() const
{
	//LAN IP からのアクセスのみ許可します。
	struct sockaddr_in peer;
	socklen_t peer_len = sizeof(peer);
	getpeername(this->ConnectSocket->native(), (struct sockaddr*)&peer, &peer_len);
#if _MSC_VER
	return peer.sin_addr.S_un.S_addr;
#else
	return peer.sin_addr.s_addr;
#endif
}

bool HttpWorker::HttpTrans()
{
//	ASSERT( this->ConnectSocket->is_open() );
	boost::asio::streambuf response_body;
	boost::system::error_code error ;
	//とりあえず今はこれでいい。タイムアウトの設定ができていないので、DoSに弱い。
	//あとでboost捨てよう。自前に置き換えてしまおう。
	boost::asio::read_until(*this->ConnectSocket, response_body,"\r\n\r\n" ,error );
	if (error)
	{
		NOTIFYLOG("XXX\t\t\tconnectionがタイムアウトしました");
		return false;
	}

	//ヘッダーパース
	const char* requestdata = boost::asio::buffer_cast<const char*>(response_body.data());
	int size = response_body.size();
	XLHttpHeader httpHeaders;
	if ( ! httpHeaders.Parse( requestdata ,size) )
	{
		HTTP500();
		NOTIFYLOG("500\t\t\tヘッダーをパース出来ません");
		return false;
	}

	//パスはよく使うので変数に入れるよ
	const std::string path = httpHeaders.getRequestPath();
	if ( strstr(path.c_str() , "/../") ) 
	{//..がくることはないっ
		HTTP403();
		NOTIFYLOG("403\t" << path << "\t\tパスに .. が入ってはいけません");
		return false;
	}

	//コンテンツを受け取る.
	const auto header = httpHeaders.getHeaderPointer();
	if (httpHeaders.getRequestMethod() == "POST")
	{
		std::map<std::string,std::string>::const_iterator it;
		if ( (it = header->find("content-length")) == header->end() )
		{
			HTTP403();
			NOTIFYLOG("403\t" << path << "\t\tPOSTリクエストにはcontent-lengthがヘッダーが必要です");
		}
		else
		{
			unsigned int contentSize = atoi(it->second.c_str());
			while(response_body.size() < httpHeaders.getHeaderSize() + contentSize )
			{
				boost::asio::read(*this->ConnectSocket, response_body
					,boost::asio::transfer_at_least(1) ,error );
				if (error)
				{
					break ;
				}
			}

			if (response_body.size() < httpHeaders.getHeaderSize() + contentSize)
			{
				HTTP403();
				NOTIFYLOG("403\t" << path << "\t\tPOSTリクエストでcontent-lengthに指定された値よりデータのほうが少ないです");
				return false;
			}
			requestdata = boost::asio::buffer_cast<const char*>(response_body.data());
			httpHeaders.PostParse(requestdata+httpHeaders.getHeaderSize() ,contentSize );
		}
	}
	//http/1.0 keep aliveできかな？
	bool isKeepAlive = false;
	if ( "keep-alive" == httpHeaders.getAt("connection") )
	{
		isKeepAlive = true;
	}
	

	//host を見る
	const std::string host = httpHeaders.getHostHeader();
	//実ファイルがあって、アクセスが許可されているならば、そいつを優先して転送する
	const std::string realpath = MainWindow::m()->GetConfigBasePath("webroot" + path);

	//存在してアクセスが許可されている拡張子か？
	if ( XLFileUtil::Exist(realpath) )
	{
		std::string mime = MainWindow::m()->Httpd.getAllowExtAndMime( XLStringUtil::strtolower(XLStringUtil::baseext_nodot(path)) );
		if (!mime.empty())
		{//実ファイルを転送する
			HTTP200SendFileContent(realpath,mime,*header,isKeepAlive);
			return isKeepAlive;
		}
	}


	//処理系に投げる
	std::string responsString;
	WEBSERVER_RESULT_TYPE result = WEBSERVER_RESULT_TYPE_NOT_FOUND;

	//CommanderWab処理系に投げる
	if (!  MainWindow::m()->CommanderWeb.WebAccess(path,httpHeaders,this->getRemoteIP(),&result,&responsString) )
	{//存在しないので 404
			HTTP404();
			NOTIFYLOG("404\t" << path << "\t" << "" << "\tCommanderWab処理系に存在しないパスです");
			return false;
	}

	switch(result)
	{
	case WEBSERVER_RESULT_TYPE_OK_HTML:
		HTTP200(responsString,"Content-type: text/html; charset=utf-8\r\n",*header,isKeepAlive);
		return false;
	case WEBSERVER_RESULT_TYPE_OK_JSON:
		HTTP200(responsString,"Content-Type: application/json; charset=utf-8\r\n",*header,isKeepAlive);
		return false;
	case WEBSERVER_RESULT_TYPE_OK_JSONP:
		{
			const auto request = httpHeaders.getRequest();
			responsString = mapfind(request,"callback","callback") + "(" +  responsString + ");";
			HTTP200(responsString,"Content-Type: text/javascript; charset=utf-8\r\n",*header,isKeepAlive);
		}
		return false;
	case WEBSERVER_RESULT_TYPE_OK_DATA:
		HTTP200(responsString,"",*header,isKeepAlive);
		return false;
	case WEBSERVER_RESULT_TYPE_TRASMITFILE:
		{
			responsString = XLStringUtil::pathseparator(responsString);
			//存在してアクセスが許可されている拡張子か？
			if (! XLFileUtil::Exist(responsString) )
			{
				HTTP404();
				NOTIFYLOG("404\t" << path << "\t" << "" << "\tコンテンツ転送(transmit)しようとしたがファイルがありません。" << responsString);
				return false;
			}
			const std::string mime = MainWindow::m()->Httpd.getAllowExtAndMime( XLStringUtil::strtolower(XLStringUtil::baseext_nodot(responsString)) );
			if (mime.empty())
			{
				HTTP403();
				NOTIFYLOG("403\t" << path << "\t" << "" << "\tコンテンツ転送(transmit)しようとしたが許可されていないmimeでした。" << responsString);
				return false;
			}
			//実ファイルを転送する
			HTTP200SendFileContent(responsString,mime,*header,isKeepAlive);
		}
		break;
	case WEBSERVER_RESULT_TYPE_ERROR:
		HTTP500();
		return false;
	case WEBSERVER_RESULT_TYPE_LOCATION:
		HTTP302(responsString);
		return false;
	case WEBSERVER_RESULT_TYPE_NOT_FOUND:
		HTTP404();
		return false;
	case WEBSERVER_RESULT_TYPE_FORBIDDEN:
		HTTP403(responsString);
		return false;
	default:
		return false;
	};

	return isKeepAlive;
}

//////////////////////////////////////////////////////////////////////
// 構築/消滅
//////////////////////////////////////////////////////////////////////
HttpServer::HttpServer()
{
	this->Thread = NULL;
	this->StopFlag = false;
	this->Acceptor = NULL;
}
HttpServer::~HttpServer()
{
	stop();
}
void HttpServer::Create()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	this->Webroot = MainWindow::m()->Config.Get("httpd__webroot", MainWindow::m()->GetConfigBasePath("webroot") );
	this->Port = MainWindow::m()->Config.GetInt("httpd__port",15551);
	this->AllowExtAndMime = MainWindow::m()->Config.FindGetsToMap("httpd__allowext_",true); 
	this->SessionDir = MainWindow::m()->Config.Get("httpd__sessiondir",MainWindow::m()->GetConfigBasePath("session"));
	int threadcount = MainWindow::m()->Config.GetInt("httpd__workerthread",2);

	//終了時にこれをcloseしてやらんと無限ループるみたい。めんどい。
	this->Acceptor = new boost::asio::ip::tcp::acceptor(  this->AcceptorIOService
		 , boost::asio::ip::tcp::endpoint(boost::asio::ip::tcp::v4(), this->Port)
	);
	//acceptスレッド作成.
	this->Thread = new boost::thread([=](){
		try
		{
			this->acceptThreadMain(threadcount); 
		}
		catch(XLException & e)
		{
			ERRORLOG(e.getFullErrorMessage());
			MainWindow::m()->Shutdown(EXITCODE_LEVEL_ERROR,true); //危険なので強制終了
		}
	});
	
}

void HttpServer::stop()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	this->StopFlag = true;

	if (this->Thread)
	{
		this->AcceptorIOService.stop();
		this->Acceptor->close();
		this->Thread->join();

		delete this->Thread;
		this->Thread = NULL;

		delete this->Acceptor;
		this->Acceptor = NULL;
	}

}

void HttpServer::acceptThreadMain(int threadcount)
{
	boost::asio::io_service ioWorker;
	boost::asio::io_service::work w( ioWorker );

	boost::thread_group tg;
	for( int i=0; i<threadcount; ++i )
	{
		tg.create_thread( boost::bind( &boost::asio::io_service::run, &ioWorker ) );
	}

	while (!this->StopFlag)
	{
		//accept のエラーをチェックしていないで、 終了時の close()すると例外で死ぬ。

		boost::system::error_code ec;
		// 接続待ち
		boost::asio::ip::tcp::socket* socket = new boost::asio::ip::tcp::socket(this->AcceptorIOService);
		this->Acceptor->accept( *socket ,ec);
		if (this->StopFlag)
		{
			delete socket;
			break;
		}

		//ワーカースレッドになげる.
		ioWorker.post([=](){ HttpWorker w(socket); w(); delete socket; });
	}
	ioWorker.stop();
	tg.join_all();
}

std::string HttpServer::GetUUID() const
{
	return MainWindow::m()->Config.Get("httpd__uuid", "" );
}

std::string HttpServer::getAllowExtAndMime(const std::string& ext) const
{
	auto it = this->AllowExtAndMime.find(ext);
	if (it == this->AllowExtAndMime.end() )
	{
		return "";
	}
	return it->second;
}

//続きはwebで
std::string HttpServer::getWebURL(const std::string& path) const
{
	std::string url = this->getServerTop() ;

	if (path.empty())        url += "/";
	else if (path[0] != '/') url += "/" + path;
	else url += path;

	return url;
}

//サーバのトップアドレス.
std::string HttpServer::getServerTop() const
{
	return MainWindow::m()->Config.Get("httpd__url", "http://127.0.0.1:" + num2str(this->Port ) );
}
