﻿// HttpServer.h: HttpServer クラスのインターフェイス
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_HttpServer_H__1477FE93_D7A8_4F29_A369_60E33C71B2B7__INCLUDED_)
#define AFX_HttpServer_H__1477FE93_D7A8_4F29_A369_60E33C71B2B7__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "XLHttpHeader.h"

enum WEBSERVER_RESULT_TYPE
{
	 WEBSERVER_RESULT_TYPE_OK_HTML
	,WEBSERVER_RESULT_TYPE_OK_JSON
	,WEBSERVER_RESULT_TYPE_OK_JSONP
	,WEBSERVER_RESULT_TYPE_OK_DATA
	,WEBSERVER_RESULT_TYPE_TRASMITFILE
	,WEBSERVER_RESULT_TYPE_ERROR
	,WEBSERVER_RESULT_TYPE_NOT_FOUND
	,WEBSERVER_RESULT_TYPE_LOCATION
	,WEBSERVER_RESULT_TYPE_FORBIDDEN
};
class HttpLongPoolingInterface
{
public:
	HttpLongPoolingInterface(){};
	virtual ~HttpLongPoolingInterface(){};
	virtual bool Pooling(int count);
	virtual std::string getFulltext() const;
};


class HttpWorker
{
	
	boost::asio::ip::tcp::socket* ConnectSocket;

public:
	HttpWorker( boost::asio::ip::tcp::socket* socket);
	virtual ~HttpWorker();

	void operator()();
private:
	void HTTP500();
	void HTTP404();
	void HTTP403();
	void HTTP403(const std::string& message);
	void HTTP302(const std::string& url);
	void HTTP200(const std::string& contents,const std::string& appendheaders,const std::map<std::string,std::string>& header,bool isKeepAlive);
	void HTTP200SendFileContent(const std::string& urlpath,const std::string& mime,const std::map<std::string,std::string>& header,bool isKeepAlive);
	bool HttpTrans();

	unsigned long getRemoteIP() const;

};

class HttpServer
{
    boost::thread*					Thread;
    boost::thread*					UPNPThread;
public:

	HttpServer();
	virtual ~HttpServer();
	void Create();
	void stop();

	std::string getAllowExtAndMime(const std::string& ext) const;
	std::string getWebURL(const std::string& path) const;

	//サーバのトップアドレス.
	std::string getServerTop() const;

	std::string getWebroot() const
	{
		return this->Webroot;
	}

	std::string GetUUID() const;

private:
	void acceptThreadMain(int threadcount);

	boost::asio::io_service AcceptorIOService;
	boost::asio::ip::tcp::acceptor* Acceptor;

	bool StopFlag;
	std::string Webroot;
	int Port;
	std::string SessionDir;

	std::map<std::string,std::string> AllowExtAndMime;
	mutable boost::mutex lock;
};


#endif // !defined(AFX_HttpServer_H__1477FE93_D7A8_4F29_A369_60E33C71B2B7__INCLUDED_)
