﻿#include "common.h"
#include "Logger.h"
#include "XLStringUtil.h"
#include "XLFileUtil.h"

Logger::Logger()
{
	this->fp = NULL;
	this->level = LOG_LEVEL_DEBUG;
	this->canStdout = true;
	this->rotateSize = 1024*1024*1; //1MB 
}

Logger::~Logger()
{
	Flush();
	if (this->fp)
	{
		fclose(this->fp);
		this->fp = NULL;
	}
}

void Logger::Create(const std::string& filename,unsigned int logBufferByte,unsigned int logRotateByte)
{
	this->rotateSize = logRotateByte;

	const auto _filename = XLStringUtil::pathseparator(filename);
	this->fp = fopen(_filename.c_str() , "ab");
	if (!this->fp)
	{
		throw XLException("ログファイル" + _filename + "が開けません");
	}
	this->fp_buffer.resize(logBufferByte);
	setbuf(this->fp,&this->fp_buffer[0]);

	this->filename = _filename;
}

void Logger::Flush()
{
	boost::unique_lock<boost::mutex> al(this->lock);	
	if (this->fp)
	{
		fflush(this->fp);
	}
}

void Logger::Write(const std::string& message)
{
	WriteLow( XLStringUtil::timetostr(time(NULL),"%Y/%m/%d %H:%M:%S" ) + "\t" + message + "\r\n" );
}

void Logger::WriteLow(const std::string& log)
{
	if (this->canStdout)
	{
#if _MSC_VER
		HANDLE handle = GetStdHandle(STD_OUTPUT_HANDLE);

		//本文 長すぎる文章を一気に書くと表示されないので、一行づつくぎって表示する.
		std::list<std::string> l = XLStringUtil::split("\r\n",log);
		for(std::list<std::string>::iterator it = l.begin() ; it != l.end() ; ++it)
		{
			WriteConsole(handle , it->c_str() , it->size(), NULL, NULL);
			WriteConsole(handle , "\r\n" , 2 , NULL, NULL);
		}
#else
		puts(log.c_str());
#endif
	}

	boost::unique_lock<boost::mutex> al(this->lock);	
	if (!this->fp)
	{
		return ;
	}
	fwrite( log.c_str(), sizeof(char) , log.size() , this->fp );

	//ローテートサイズを超えていたら新規にスタートする
	unsigned long pos = ftell(this->fp);
	if ( pos >  this->rotateSize )
	{
		rotateLow();
	}
}

std::string Logger::tail(int lines)
{
	Flush();
	return XLFileUtil::tail(this->filename,lines);
}
//ローテート(public)
void Logger::rotate()
{
	boost::unique_lock<boost::mutex> al(this->lock);
	rotateLow();
}

//ローテートの実装部分(private) 必ず lockして読んでね
void Logger::rotateLow()
{
	if (!this->fp) return ;
	fclose(this->fp);
	//常に1世代保存
	XLFileUtil::copy( getFilename() , getFilenameBack() );
	XLFileUtil::del( getFilename() );
	Create( getFilename() ,this->fp_buffer.size() , this->rotateSize );
}
