﻿#include "common.h"
#include "Config.h"
#include "XLStringUtil.h"
#include "XLFileUtil.h"
#include "MainWindow.h"

#if _MSC_VER
	#include "resource1.h"
	#include "methodcallback.h"

	const char* HIDDEN_WINDOW_NAME = "FHCCommander_Hidden";
#endif //_MSC_VER
const int WM_THREAD_CALLBACK		 = WM_USER + 1;
const int WM_USER_TASKTRAY_CLICK	 = WM_USER + 2;
const int WM_TASKBAR_CREATED		 = WM_USER + 3;


MainWindow* MainWindow::SingletonPoolMainWindow = NULL;

MainWindow::MainWindow()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません
	ASSERT(MainWindow::SingletonPoolMainWindow == NULL); //シングルトンです
	MainWindow::SingletonPoolMainWindow = this; //まーいいだろう。 (スレッド競合は気にせんでいい)

	this->WindowInstance = NULL;
	this->MainWindowHandle = NULL;
	this->ConfigBaseDirectory = XLStringUtil::pathcombine( XLFileUtil::pwd() , "config" );

	//config読み込み
	this->Config.Create( GetConfigBasePath("config.conf") );

	//loggerの準備
	this->Logger.Create( 
		 this->Config.Get("log__filename", GetConfigBasePath("log") + "/log.txt") 
		,this->Config.GetInt("log__bufferbyte",65535)
		,this->Config.GetInt("log__rotatebyte",1024*1024*1)
		);
	this->Logger.SetCanStdout(this->IsOpenConsole);
	this->Logger.SetLogLevel( (LOG_LEVEL) (this->Config.GetInt("log__level",(int)LOG_LEVEL_DEBUG)) );
}
MainWindow::~MainWindow()
{
}

bool MainWindow::Create(bool isdebug)
{
	this->MainThreadID = ::boost::this_thread::get_id();
	this->StopFlag = false;
	this->ExitLevel = EXITCODE_LEVEL_ERROR;

#if _MSC_VER
	this->WindowInstance = GetModuleHandle(0);
	WNDCLASS myProg;
	myProg.style            =CS_HREDRAW | CS_VREDRAW;
	myProg.lpfnWndProc      =methodcallback::registstdcall<struct _mainwindow_winproc,WNDPROC>(this, &MainWindow::WndProc);
	myProg.cbClsExtra       =0;
	myProg.cbWndExtra       =0;
	myProg.hInstance        =this->WindowInstance ;
	myProg.hIcon            =NULL;
	myProg.hCursor          =LoadCursor(NULL, IDC_ARROW);
	myProg.hbrBackground    = (HBRUSH)GetStockObject(WHITE_BRUSH);
	myProg.lpszMenuName     =NULL;
	myProg.lpszClassName    = HIDDEN_WINDOW_NAME;
	if (!RegisterClass(&myProg))
	{
		throw XLException("windowクラスの登録に失敗!");
	}

	this->MainWindowHandle = CreateWindow(HIDDEN_WINDOW_NAME
		,HIDDEN_WINDOW_NAME
		,WS_OVERLAPPEDWINDOW
		,CW_USEDEFAULT
		,CW_USEDEFAULT
		,500
		,400
		,NULL
		,NULL
		,this->WindowInstance 
		,NULL);

//	ShowWindow(this->MainWindowHandle,SW_SHOW);
//	UpdateWindow(this->MainWindowHandle);
	this->MCIPlayStatus = 0;
#else
	this->MainWindowHandle = NULL;
	this->WindowInstance = NULL;
#endif

	if (! OnInit() )
	{
		ERRORLOG("OnInit に失敗しました.");
		return false;
	}

	//見つけた!
	std::string weburl = this->Httpd.getWebURL("/local/");
#if _MSC_VER
	ShellExecute(NULL,NULL,weburl.c_str() , NULL,NULL,0);
#else
#endif
	return true;
}

//既に動いていないかチェックする.
//(static メソッド)
bool MainWindow::checkAlreadyRunning() 
{
#if _MSC_VER
	//mutexでもいいけど、ここではmessageを送信したいので
	//main windowがあるかどうかを検索する.
	HWND hwnd = ::GetTopWindow(::GetDesktopWindow());
	do
	{
		TCHAR title[ MAX_PATH ]= {0};
		::GetWindowText( hwnd, title, sizeof(title) );

		if ( lstrcmp(title , HIDDEN_WINDOW_NAME ) != 0 )
		{
			continue;
		}
		//見つけた!
		PostMessage(hwnd , WM_COMMAND, (WPARAM)ID_MENU_WEB, 0); 
		//既に動いてます!
		return true;
	}
	while( (hwnd = ::GetNextWindow(hwnd,GW_HWNDNEXT)) !=NULL );
#else
#endif
	//まだ動いていません!
	return false;
}

EXITCODE_LEVEL MainWindow::BlockMessageLoop()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

#if _MSC_VER
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
#else
	while(!this->StopFlag)
	{
		{
			boost::unique_lock<boost::mutex> al(this->Lock);
			if (this->Queue.empty())
			{
				this->queue_wait.wait(al);
			}
		}
		//寝起きかもしれないので終了条件の確認.
		if (this->StopFlag)
		{
			return 0;
		}
		//メッセージをすべて再生
		RunMessagePump();
	}
#endif
	return this->ExitLevel;
}
LRESULT MainWindow::WndProc(HWND hWnd, unsigned int msg, WPARAM wParam, LPARAM lParam)
{
	//メッセージによる分岐
	switch (msg) {
		case WM_CREATE:
			if (! OnCreate() )
			{
				return -1;
			}
			break;
		case WM_DESTROY:
			OnDestory((EXITCODE_LEVEL)wParam);
			break;
		case WM_THREAD_CALLBACK:
			if (wParam && lParam)
			{
				((_THREAD_CALLBACK)wParam)(lParam);
			}
			break;
#if _MSC_VER
		case WM_USER_TASKTRAY_CLICK:
			OnTaskTrayClick(wParam, lParam);
			break;
		case WM_TASKBAR_CREATED:
			OnTaskbarCreated(wParam, lParam);
			break;
		case WM_COMMAND:
			OnCommand(wParam, lParam);
			break;
		case MM_MCINOTIFY:
			this->MCIPlayStatus = 0;
			break;
		default:
			return(DefWindowProc(hWnd, msg, wParam, lParam));
#endif
	}
	return (0L);
}

void MainWindow::OnCommand(WPARAM wParam, LPARAM lParam)
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません
#if _MSC_VER
	if (wParam == ID_MENU_QUIT)
	{//終了
		SendMessage(this->MainWindowHandle,WM_DESTROY,0,0);
	}
	else if (wParam == ID_MENU_WEB)
	{//web画面を開く
		std::string weburl = this->Httpd.getWebURL("/local/");
		ShellExecute(NULL,NULL,weburl.c_str() , NULL,NULL,0);
	}
	else if (wParam == ID_MENU_ABOUT)
	{//about画面を開く
		std::string weburl = this->Httpd.getWebURL("/about/");
		ShellExecute(NULL,NULL,weburl.c_str() , NULL,NULL,0);
	}
#endif
}

//windows依存部
#if _MSC_VER

//通知領域に追加
bool MainWindow::AddNotifyIcon()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	//タスクトレイにアイコンを出す
	this->NotifyIcon.cbSize = sizeof(NOTIFYICONDATA);
	this->NotifyIcon.uID = 0;
	this->NotifyIcon.hWnd = this->MainWindowHandle;
	this->NotifyIcon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	this->NotifyIcon.hIcon = LoadIcon(this->WindowInstance,MAKEINTRESOURCE(IDI_ICON1));
	this->NotifyIcon.uCallbackMessage = WM_USER_TASKTRAY_CLICK;
	lstrcpyA( this->NotifyIcon.szTip, "FHCコマンダー" );

	// アイコンを登録できるかエラーになるまで繰り返す
	while (true)
	{
		if (Shell_NotifyIcon(NIM_ADD, &this->NotifyIcon))
		{
			// 登録できたら終わり
			break;
		}
		else
		{
			// タイムアウトかどうか調べる
			if (::GetLastError() != ERROR_TIMEOUT)
			{
				// アイコン登録エラー
				return false;
			}

			// 登録できていないことを確認する
			if (Shell_NotifyIcon(NIM_MODIFY, &this->NotifyIcon))
			{
				// 登録できていた
				break;
			}
			else
			{
				// 登録できていなかった
				SecSleep(1);
			}
		}
	}
	return true;
}

//タスクトレイからアイコンを消す
bool MainWindow::DelNotifyIcon()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	::Shell_NotifyIcon( NIM_DELETE, &this->NotifyIcon );
	return true;
}

//タスクトレイのアイコンがクリックされたとき
LRESULT MainWindow::OnTaskTrayClick(WPARAM wParam, LPARAM lParam)
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	// カーソルの現在位置を取得
	POINT point;
	::GetCursorPos(&point);

	switch(lParam) 
	{
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		SetForegroundWindow(this->MainWindowHandle);
		{
			HMENU menu = LoadMenu( this->WindowInstance , MAKEINTRESOURCE(IDR_MENU2) );
			if (!menu)
			{
				break;
			}
			HMENU menu2 = GetSubMenu(menu,0);
			if (!menu2)
			{
				break;
			}
			TrackPopupMenu(menu2,TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y,0,this->MainWindowHandle,NULL);
			PostMessage(this->MainWindowHandle,WM_NULL,0,0);
		}
	}
	return 0;
}

//シェルが死んだ時に飛んでくるメッセージ
LRESULT MainWindow::OnTaskbarCreated(WPARAM wParam, LPARAM lParam)
{
	this->AddNotifyIcon();
	return 0;
}


//MCIで音楽を再生する。再生が終わるまで待つ
void MainWindow::PlayMCISync(const std::string& filename )
{
	MCI_OPEN_PARMS openParms = {0};
	MCI_PLAY_PARMS playParms = {0};
	MCIERROR  r;

	const std::string ext = XLStringUtil::baseext_nodot(filename);
	if (ext == "mp3")
	{
		openParms.lpstrElementName = filename.c_str();
		openParms.lpstrDeviceType = (LPTSTR)"MPEGVideo";

		r = mciSendCommand(0, MCI_OPEN,
			MCI_OPEN_TYPE | MCI_OPEN_ELEMENT,
			(DWORD)(LPVOID)&openParms);
		if (r != 0)
		{
			throw XLException("音楽を再生できません MCI_OPENに失敗(" + num2str(r) + ") filename:" + filename);
		}

	}
	else if (ext == "wav")
	{
		openParms.lpstrElementName = filename.c_str();
		openParms.lpstrDeviceType = (LPTSTR)MCI_DEVTYPE_WAVEFORM_AUDIO;

		r = mciSendCommand(0, MCI_OPEN,
			MCI_OPEN_TYPE | MCI_OPEN_TYPE_ID | MCI_OPEN_ELEMENT,
			(DWORD)(LPVOID)&openParms);
		if (r != 0)
		{
			throw XLException("音楽を再生できません MCI_OPENに失敗(" + num2str(r) + ") filename:" + filename);
		}
	}
	else
	{
		throw XLException("サポートしていないファイル形式です。 filename:" + filename);
	}

	//再生中だと終わるまで待つ
	while (this->MCIPlayStatus == 1)
	{
		SecSleep(1); 
	}
	this->MCIPlayStatus = 1;
	
	playParms.dwFrom = 0;
	playParms.dwCallback = (DWORD_PTR) this->MainWindowHandle;
	r = mciSendCommand(openParms.wDeviceID, MCI_PLAY , MCI_NOTIFY | MCI_WAIT ,
				(DWORD)(LPVOID)&playParms);
	if (r != 0)
	{
		throw XLException("音楽を再生できません MCI_PLAYに失敗(" + num2str(r) + ") filename:" + filename);
	}

	if (::boost::this_thread::get_id() == this->MainThreadID) 
	{
		ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドですね

		MSG msg;
		while (GetMessage(&msg, NULL, 0, 0)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);

			if (this->MCIPlayStatus == 0) break;
		}
	}
	else
	{
		while(1) 
		{
			SecSleep(1); 
			if (this->MCIPlayStatus == 0) break;
		}
	}

	MCI_GENERIC_PARMS genericParam = {0};
	r = mciSendCommand(openParms.wDeviceID, MCI_CLOSE , 0 ,
				(DWORD)(LPVOID)&genericParam);

	return ;
}


#else
//linux依存部

//メッセージを処理する
void MainWindow::RunMessagePump()
{
	boost::unique_lock<boost::mutex> al(this->Lock);
	for(auto it = this->Queue.begin() ; it != this->Queue.end() ; ++it )
	{
		LRESULT r = WndProc(0, it->message,it->wParam,it->lParam);
		if (it->result && it->updatewait)
		{
			*(it->result) = r;
			*(it->updatewait) = true;
		}
	}
	this->Queue.clear();
}
LRESULT MainWindow::SendMessage(HWND dummyWindowHandle,unsigned int message,WPARAM wParam,LPARAM lParam)
{
	int result = 0;
	bool updatewait = false;
	struct Message m = {0};
	m.message = message;
	m.wParam = wParam;
	m.lParam = lParam;
	m.result = &result;
	m.updatewait = &updatewait;
	
	//仕事を積む
	{
		boost::unique_lock<boost::mutex> al(this->Lock);
		this->Queue.push_back(m);
	}
	//スレッドwakeup
	{
		this->queue_wait.notify_all();
	}

	if (this->MainThreadID == ::boost::this_thread::get_id())
	{
		//メインスレッドであれば、メッセージポンプを回す。 そうじゃないと詰まってしまう。
		RunMessagePump();
	}
	
	//値を取得する(値を取得するまでブロック).
	while(!updatewait)
	{
		sleep(0);
	}

	return result;
}
void MainWindow::PostMessage(HWND dummyWindowHandle,unsigned int message,WPARAM wParam,LPARAM lParam)
{
	struct Message m = {0};
	m.message = message;
	m.wParam = wParam;
	m.lParam = lParam;
	m.result = NULL;
	m.updatewait = NULL;

	//仕事を積む
	{
		boost::unique_lock<boost::mutex> al(this->Lock);
		this->Queue.push_back(m);
	}
	//スレッドwakeup
	{
		this->queue_wait.notify_all();
	}
}


#endif

//終了時
void MainWindow::OnDestory(EXITCODE_LEVEL level)
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

	//停止フラグを立てる.
	this->StopFlag = true;
	//終了レベルを記録
	this->ExitLevel = level;

#if _MSC_VER
	//今あるメッセージ・キューを空にする.
	{
		MSG msg;
		while(PeekMessage (&msg,NULL,0,0,PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	DEBUGLOG("メッセージループ完了。");
#endif

	SyncFiles();

#if _MSC_VER
	//タスクトレイからアイコンを消す.
	DelNotifyIcon();
	//自滅用のメッセージを送信.
	PostQuitMessage(0);
#endif

	//ウィンドウハンドルのクリア.
	this->MainWindowHandle = NULL;
}

//作成された時
bool MainWindow::OnCreate()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません
	return true;
}



//Popupメッセージを出す.
void MainWindow::PopupMessage(const std::string & title,const std::string & message )
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

#if _MSC_VER
	//バルーンを出す.
	NOTIFYICONDATA	sData;
	ZeroMemory(&sData, sizeof(NOTIFYICONDATA));
	sData.cbSize		= sizeof(NOTIFYICONDATA);
	sData.uFlags		= NIF_INFO;
	sData.hWnd			= this->MainWindowHandle;
	sData.dwInfoFlags	= NIIF_INFO;
	sData.uTimeout		= 2000;

	lstrcpyA(sData.szInfoTitle,title.c_str() );
	lstrcpyA(sData.szInfo,  message.c_str() );

	Shell_NotifyIcon(NIM_MODIFY,&sData);
#endif
}

//初期化を行う所
bool MainWindow::OnInit()
{
	ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

#if _MSC_VER
	bool ret = this->AddNotifyIcon();
	if (!ret)
	{
		return false;
	}
#endif
	this->CommanderWeb.Create();

	//全体的の初期化を行います。
	this->Httpd.Create();

	return true;
}


//バッファリングしているものをsyncする
void MainWindow::SyncFiles()
{
	this->Config.overrideSave();
}


//終了させる.
void MainWindow::Shutdown(EXITCODE_LEVEL level,bool isForce)
{
	//ファイルをすべて保存する
	SyncFiles();

	if (isForce)
	{//今すぐ叩き殺す!! メモリは OSが開放してくれるでしょうwww
		exit((int)level);
	}
	else
	{//自滅するメッセージを送信
		PostMessage(
			 this->MainWindowHandle
			,WM_DESTROY
			,(WPARAM)level
			,(LPARAM)isForce
		);
	}
}


void MainWindow::SyncInvoke(std::function<void (void) > func)
{
	//停止フラグが有効な場合は無視する。
	if (this->StopFlag)
	{
		return ;
	}

	if (::boost::this_thread::get_id() == this->MainThreadID) 
	{
		ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドですね
		//即実行
		func();
		return ;
	}

	struct callbackClass
	{
		callbackClass(std::function<void (void) > func) : func(func)
		{
			
		}
		static void callback(int p)
		{
			ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

			callbackClass* _this = (callbackClass*)p;
			_this->func();
		}

		std::function<void (void) > func;
	};

	callbackClass call(func);
	SendMessage(
		 this->MainWindowHandle
		,WM_THREAD_CALLBACK
		,(WPARAM)callbackClass::callback
		,(LPARAM)&call
	);
}

void MainWindow::AsyncInvoke(std::function<void (void) > func)
{
	//停止フラグが有効な場合は無視する。
	if (this->StopFlag)
	{
		return ;
	}

	struct callbackClass
	{
		callbackClass(std::function<void (void) > func) : func(func)
		{
			
		}
		static void callback(int p)
		{
			ASSERT_IS_MAIN_THREAD_RUNNING(); //メインスレッドでしか動きません

			callbackClass* _this = (callbackClass*)p;
			_this->func();

			delete _this;
		}

		std::function<void (void) > func;
	};
	callbackClass *call = new callbackClass(func);
	PostMessage(
		 this->MainWindowHandle
		,WM_THREAD_CALLBACK
		,(WPARAM)callbackClass::callback
		,(LPARAM)call
	);
}

void MainWindow::SyncInvokePopupMessage(const std::string& title,const std::string& message)
{
	//停止フラグが有効な場合は無視する。
	if (this->StopFlag)
	{
		return ;
	}

	this->SyncInvoke( [=](){
		this->PopupMessage( title, message) ;
	});
}


//config の中のパスを作る
std::string MainWindow::GetConfigBasePath(const std::string& path) const
{
	return XLStringUtil::pathcombine( this->ConfigBaseDirectory , path);
}

