﻿#pragma once

#include "common.h"
#include "Logger.h"
#include "Config.h"
#include "HttpServer.h"
#include "CommanderWeb.h"

enum EXITCODE_LEVEL
{
	 EXITCODE_LEVEL_REBOOT   = 10 //プログラムを終了して、OSごと再起動します
	,EXITCODE_LEVEL_SHUTDOWN = 20 //プログラムを終了して、システムをシャットダウンします
	,EXITCODE_LEVEL_ERROR    = 5  //プログラムはエラーで死にました。ザオリクします。 
	,EXITCODE_LEVEL_NORMALEND= 0  //プログラムは正常に終了しました。
};

class MainWindow
{
public:
	MainWindow();
	virtual ~MainWindow();

	bool Create(bool isdebug);
	void Shutdown(EXITCODE_LEVEL level,bool isForce);

	EXITCODE_LEVEL BlockMessageLoop();

	static bool checkAlreadyRunning() ;
	
	void SyncInvokePopupMessage(const std::string& title,const std::string& message);

	//初期化を行う所
	typedef void (*_THREAD_CALLBACK)(int p);
	//メインスレッドコールバック 関数オブジェクト
	void SyncInvoke(std::function<void (void) > func);
	//メインスレッドで後で実行するコールバック
	void AsyncInvoke(std::function<void (void) > func);
	//config の中のパスを作る
	std::string GetConfigBasePath(const std::string& path) const;
	//バッファリングしているものをsyncする
	void SyncFiles();
	
	static MainWindow* m()
	{
		return MainWindow::SingletonPoolMainWindow;
	}

#if _MSC_VER
	//MCIで音楽を再生する。再生が終わるまで待つ
	void PlayMCISync(const std::string& filename );
#endif
private:
	virtual LRESULT WndProc(HWND hWnd, unsigned int msg, WPARAM wParam, LPARAM lParam);
	bool OnInit();

	//終了時
	void OnDestory(EXITCODE_LEVEL level);
	//作成された時
	bool OnCreate();
	//Popupメッセージを出す.
	void PopupMessage(const std::string & title,const std::string & message);
	void OnCommand(WPARAM wParam, LPARAM lParam);


#if _MSC_VER
//windows依存部
	//通知領域に追加
	bool AddNotifyIcon();
	//タスクトレイからアイコンを消す
	bool DelNotifyIcon();
	//タスクトレイのアイコンがクリックされたとき
	LRESULT OnTaskTrayClick(WPARAM wParam, LPARAM lParam);
	//シェルが死んだ時に飛んでくるメッセージ
	LRESULT OnTaskbarCreated(WPARAM wParam, LPARAM lParam);

#else
//linux依存部
	//windowsの SendMessage相当のことをする
	LRESULT SendMessage(HWND dummyWindowHandle,unsigned int message,WPARAM wParam,LPARAM lParam);
	//windowsの PostMessage相当のことをする
	void PostMessage(HWND dummyWindowHandle,unsigned int message,WPARAM wParam,LPARAM lParam);
	//メッセージを処理する
	void RunMessagePump();
#endif


public:
	class Config					Config;
	class Logger					Logger;
	class HttpServer				Httpd;
	class CommanderWeb				CommanderWeb;

private:
	::boost::thread::id MainThreadID;
	bool IsOpenConsole;
	bool StopFlag;
	HINSTANCE WindowInstance;
	HWND MainWindowHandle;
	EXITCODE_LEVEL ExitLevel;
	std::string ConfigBaseDirectory;
	static MainWindow* SingletonPoolMainWindow;

#if _MSC_VER
//windows依存部
	NOTIFYICONDATA NotifyIcon;
	int MCIPlayStatus;
#else
//linuxでもwindowsみたいなメッセージループを擬似的に作る
	boost::mutex   Lock;
	boost::condition_variable queue_wait;
	struct Message
	{
		unsigned int	message;
		WPARAM	wParam;
		LPARAM	lParam;
		int*	result;
		bool*	updatewait;
	};
	std::list<Message> Queue;
#endif
};

class COMInit
{
public:
	COMInit()
	{
#if _MSC_VER
		::CoInitialize(NULL);
#else
#endif
	}
	virtual ~COMInit()
	{
#if _MSC_VER
		::CoUninitialize();
#else
#endif
	}
};

class WinSockInit
{
public:
	WinSockInit()
	{
#if _MSC_VER
//		WSADATA wsaData;
//		::WSAStartup(2 , &wsaData);
#else
#endif
	}
	virtual ~WinSockInit()
	{
#if _MSC_VER
//		::WSACleanup();
#else
#endif
	}
};

