﻿#include "common.h"
#include "XLSoundPlay.h"
#include "XLStringUtil.h"
#include "XLFileUtil.h"
#include "MainWindow.h"


#ifdef _MSC_VER
#pragma comment(lib,"winmm.lib")

class XLSoundPlay_MCI : public IXLSoundPlay_Interface
{
public:
	XLSoundPlay_MCI()
	{
	}
	virtual ~XLSoundPlay_MCI()
	{
	}

	virtual bool play(const std::string & filename)
	{
//		PlaySound(filename.c_str() , NULL , SND_SYNC | SND_FILENAME);
		MainWindow::m()->PlayMCISync(filename);
		return true;
	}
};
#else
#include <alsa/asoundlib.h>

class XLSoundPlay_ALSA : public IXLSoundPlay_Interface
{
public:
	XLSoundPlay_ALSA()
	{
	}
	virtual ~XLSoundPlay_ALSA()
	{
	}

	virtual bool play(const std::string & filename)
	{
		char* wavbuffer;
		unsigned long wavbuffersize;
		int sampling_rate;
		int format;
		
		const std::string ext = XLStringUtil::baseext_nodot(filename);
		
		if (ext == "mp3")
		{
//			const std::string cmd = "/usr/local/bin/mpg123 \"" + filename + "\"";
			const std::string cmd = "/usr/bin/mplayer \"" + filename + "\"";
			system(cmd.c_str() );
			return true;
		}
		else if (ext == "wav")
		{
//			LoadWav(filename,&wavbuffer,&wavbuffersize,&sampling_rate,&format);
//			PlayAlsa(wavbuffer,wavbuffersize,sampling_rate,format);
			const std::string cmd = "/usr/bin/mplayer \"" + filename + "\"";
			system(cmd.c_str() );
			return true;
		}
		else
		{
			throw XLException("サポートしていないファイル形式です。 filename:" + filename);
		}
		
	}
private:
	std::vector<char> FileBuffer;
	void LoadWav(const std::string& filename,char** wavbuffer,unsigned long* wavbuffersize , int* sampling_rate, int* format)
	{
		 if ( !  XLFileUtil::cat_b(filename , &this->FileBuffer) )
		 {
			 throw XLException("ファイル" + filename + "が読み込めません");
		 }
		 char* p = &this->FileBuffer[0];
		 if ( memcmp( p , "RIFF" , 4) != 0 )
		 {
			 throw XLException("ファイル" + filename + "は waveではありません");
		 }
//		 unsigned long riff_chunk_size          = *(unsigned long*) (p + 4);
//		 unsigned long riff_form_type           = *(unsigned long*) (p + 4 + 4);
//		 unsigned long fmt_chunk_ID             = *(unsigned long*) (p + 4 + 4 + 4);
//		 unsigned long fmt_chunk_size           = *(unsigned long*) (p + 4 + 4 + 4 + 4);
//		 unsigned short fmt_wave_format_type    = *(unsigned short*)(p + 4 + 4 + 4 + 4 + 4);
//		 unsigned short fmt_channel             = *(unsigned short*)(p + 4 + 4 + 4 + 4 + 4 + 2);
		 unsigned long fmt_samples_per_sec      = *(unsigned long*) (p + 4 + 4 + 4 + 4 + 4 + 2 + 2);
//		 unsigned long fmt_bytes_per_sec        = *(unsigned long*) (p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4);
//		 unsigned short fmt_block_size          = *(unsigned short*)(p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4);
//		 unsigned short fmt_bits_per_sample     = *(unsigned short*)(p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4 + 2);
//		 const char* data_chunk_ID              =                   (p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4 + 2 + 2);
		 *wavbuffersize                         = *(unsigned long*) (p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4 + 2 + 2 + 4);
		 *wavbuffer                             =                   (p + 4 + 4 + 4 + 4 + 4 + 2 + 2 + 4 + 4 + 2 + 2 + 4 + 4);

		 *sampling_rate = fmt_samples_per_sec;
		 *format = SND_PCM_FORMAT_S16_LE;
		 *wavbuffersize = *wavbuffersize / 2;
	}

	void PlayAlsa(const char* wavbuffer,int wavbuffersize , int sampling_rate, int format )
	{
		// 再生用PCMストリームを開く
		snd_pcm_t *stream;
		int err = snd_pcm_open( &stream, "default", SND_PCM_STREAM_PLAYBACK, 0 );
		if( err < 0 ) 
		{
			snd_pcm_close(stream);
			throw XLException(std::string("snd_pcm_open error:") + snd_strerror(err) );
		}

		err = snd_pcm_set_params ( stream, format, SND_PCM_ACCESS_RW_INTERLEAVED, 1, sampling_rate, 1, 500000 );
		if (err < 0 )
		{
			snd_pcm_close(stream);
			throw XLException(std::string("snd_pcm_set_params error:") + snd_strerror(err) );
		}
 
		err = snd_pcm_writei(stream, wavbuffer, wavbuffersize);
		if (err < 0)
		{
			err = snd_pcm_recover(stream , err, 0 );
			if( err < 0 ) 
			{
				snd_pcm_drain(stream);
				snd_pcm_close(stream);
				throw XLException(std::string("snd_pcm_writei and snd_pcm_recover error:") + snd_strerror(err) );
			}
			err = 0;
		}
		snd_pcm_drain(stream);
		snd_pcm_close(stream);
	}
};
#endif


//なんとかして音楽(wav/mp3も)を再生する.
XLSoundPlay::XLSoundPlay()
{
#ifdef _MSC_VER
	this->Object = new XLSoundPlay_MCI;
#else
	this->Object = new XLSoundPlay_ALSA;
#endif
}

XLSoundPlay::~XLSoundPlay()
{
	delete this->Object;
}

bool XLSoundPlay::play(const std::string & filename) 
{
	return this->Object->play(filename);
}

